"use strict";
/*
 * Copyright 2020 NEM
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.LinkService = void 0;
const _ = require("lodash");
const rxjs_1 = require("rxjs");
const operators_1 = require("rxjs/operators");
const symbol_sdk_1 = require("symbol-sdk");
const logger_1 = require("../logger");
const LoggerFactory_1 = require("../logger/LoggerFactory");
const BootstrapUtils_1 = require("./BootstrapUtils");
const ConfigLoader_1 = require("./ConfigLoader");
const logger = LoggerFactory_1.default.getLogger(logger_1.LogType.System);
class LinkService {
    constructor(params) {
        this.params = params;
        this.configLoader = new ConfigLoader_1.ConfigLoader();
    }
    async run(passedPresetData, passedAddresses) {
        const presetData = passedPresetData !== null && passedPresetData !== void 0 ? passedPresetData : this.configLoader.loadExistingPresetData(this.params.target);
        const addresses = passedAddresses !== null && passedAddresses !== void 0 ? passedAddresses : this.configLoader.loadExistingAddresses(this.params.target);
        const url = this.params.url.replace(/\/$/, '');
        const repositoryFactory = new symbol_sdk_1.RepositoryFactoryHttp(url);
        const currency = (await repositoryFactory.getCurrencies().toPromise()).currency;
        logger.info(`${this.params.unlink ? 'Unlinking' : 'Linking'} nodes using network url ${url}. Max Fee ${this.params.maxFee / Math.pow(10, currency.divisibility)}`);
        const generationHash = await repositoryFactory.getGenerationHash().toPromise();
        if (generationHash !== presetData.nemesisGenerationHashSeed) {
            throw new Error(`You are connecting to the wrong network. Expected generation hash is ${presetData.nemesisGenerationHashSeed} but got ${generationHash}`);
        }
        const epochAdjustment = await repositoryFactory.getEpochAdjustment().toPromise();
        const height = (await repositoryFactory.createChainRepository().getChainInfo().toPromise()).height;
        const transactionNodes = this.createTransactionsToAnnounce(epochAdjustment, height, addresses, presetData);
        if (!transactionNodes.length) {
            logger.info(`There are no transactions to announce...`);
            return;
        }
        const transactionRepository = repositoryFactory.createTransactionRepository();
        const transactionService = new symbol_sdk_1.TransactionService(transactionRepository, repositoryFactory.createReceiptRepository());
        const listener = repositoryFactory.createListener();
        await listener.open();
        const faucetUrl = presetData.faucetUrl;
        const signedTransactionObservable = rxjs_1.from(transactionNodes).pipe(operators_1.mergeMap(({ node, transactions }) => {
            if (!node.main) {
                throw new Error('CA account is required!');
            }
            const account = symbol_sdk_1.Account.createFromPrivateKey(node.main.privateKey, presetData.networkType);
            const noFundsMessage = faucetUrl
                ? `Does your node signing address have any network coin? Send some tokens to ${account.address.plain()} via ${faucetUrl}/?recipient=${account.address.plain()}`
                : `Does your node signing address have any network coin? Send some tokens to ${account.address.plain()} .`;
            return repositoryFactory
                .createAccountRepository()
                .getAccountInfo(account.address)
                .pipe(operators_1.mergeMap((a) => {
                const currencyMosaicIdHex = BootstrapUtils_1.BootstrapUtils.toHex(presetData.currencyMosaicId);
                const mosaic = a.mosaics.find((m) => BootstrapUtils_1.BootstrapUtils.toHex(m.id.toHex()) === currencyMosaicIdHex);
                if (!mosaic || mosaic.amount.compare(symbol_sdk_1.UInt64.fromUint(0)) < 1) {
                    logger.error(`Node signing account ${account.address.plain()} does not have enough currency. Mosaic id: ${currencyMosaicIdHex}. \n\n${noFundsMessage}`);
                    return rxjs_1.EMPTY;
                }
                return rxjs_1.from(transactions.map((t) => account.sign(t, generationHash)));
            }), operators_1.catchError((e) => {
                logger.error(`Node signing account ${account.address.plain()} is not valid. ${e.message}. \n\n${noFundsMessage}`);
                return rxjs_1.EMPTY;
            }));
        }));
        const announceCalls = signedTransactionObservable.pipe(operators_1.mergeMap((signedTransaction) => {
            return transactionService.announce(signedTransaction, listener).pipe(operators_1.map((completedTransaction) => {
                var _a, _b;
                const message = `Transaction ${completedTransaction.type} ${(_a = completedTransaction.transactionInfo) === null || _a === void 0 ? void 0 : _a.hash} - signer ${(_b = completedTransaction.signer) === null || _b === void 0 ? void 0 : _b.address.plain()} has been confirmed`;
                logger.info(message);
                return message;
            }), operators_1.catchError((e) => {
                const message = `Transaction ${signedTransaction.type} ${signedTransaction.hash} - signer ${signedTransaction.getSignerAddress().plain()} failed!! ` + e.message;
                logger.error(message);
                return rxjs_1.of(message);
            }));
        }));
        await announceCalls.pipe(operators_1.toArray()).toPromise();
        listener.close();
    }
    createTransactionsToAnnounce(epochAdjustment, height, addresses, presetData) {
        return _.flatMap(addresses.nodes || [])
            .filter((node) => node.main && (node.remote || node.voting || node.vrf))
            .map((node) => {
            const transactions = [];
            if (!node.main) {
                throw new Error('CA private key is required!');
            }
            const account = symbol_sdk_1.Account.createFromPrivateKey(node.main.privateKey, presetData.networkType);
            const action = this.params.unlink ? symbol_sdk_1.LinkAction.Unlink : symbol_sdk_1.LinkAction.Link;
            logger.info(`Creating transactions for node: ${node.name}, ca/main account: ${account.address.plain()}`);
            const deadline = symbol_sdk_1.Deadline.create(epochAdjustment);
            const maxFee = symbol_sdk_1.UInt64.fromUint(this.params.maxFee);
            if (node.remote) {
                logger.info(`Creating AccountKeyLinkTransaction - node: ${node.name}, signer public key: ${account.publicKey}, Remote Account public key: ${node.remote.publicKey}`);
                transactions.push(symbol_sdk_1.AccountKeyLinkTransaction.create(deadline, node.remote.publicKey, action, presetData.networkType, maxFee));
                logger.info(`Creating NodeKeyLinkTransaction - node: ${node.name}, signer public key: ${account.publicKey}, Transport/Node Account public key: ${node.transport.publicKey}`);
                transactions.push(symbol_sdk_1.NodeKeyLinkTransaction.create(deadline, node.transport.publicKey, action, presetData.networkType, maxFee));
            }
            if (node.vrf) {
                logger.info(`Creating VrfKeyLinkTransaction - node: ${node.name}, signer public key: ${account.publicKey}, VRF public key: ${node.vrf.publicKey}`);
                transactions.push(symbol_sdk_1.VrfKeyLinkTransaction.create(deadline, node.vrf.publicKey, action, presetData.networkType, maxFee));
            }
            if (node.voting) {
                logger.info(`Creating VotingKeyLinkTransaction - node: ${node.name}, signer public key: ${account.publicKey}, Voting public key: ${node.voting.publicKey}`);
                transactions.push(BootstrapUtils_1.BootstrapUtils.createVotingKeyTransaction(node.voting.publicKey, height, presetData, deadline, maxFee));
            }
            return { node, transactions };
        });
    }
}
exports.LinkService = LinkService;
LinkService.defaultParams = {
    target: BootstrapUtils_1.BootstrapUtils.defaultTargetFolder,
    url: 'http://localhost:3000',
    maxFee: 100000,
    unlink: false,
};
